from . import epics_event_pb2 as proto

#: select protocol buffer decoder for EPICS type
decoder = {
    0: proto.ScalarString,
    1: proto.ScalarShort,
    2: proto.ScalarFloat,
    3: proto.ScalarEnum,
    4: proto.ScalarByte,
    5: proto.ScalarInt,
    6: proto.ScalarDouble,
    7: proto.VectorString,
    8: proto.VectorShort,
    9: proto.VectorFloat,
    10: proto.VectorEnum,
    11: proto.VectorChar,
    12: proto.VectorInt,
    13: proto.VectorDouble,
    14: proto.V4GenericBytes,
}

#: translate EPICS types to numpy types
dtypes = {
    0: "S40",
    1: "i2",
    2: "f4",
    3: "u4",
    4: "b1",
    5: "i4",
    6: "f8",
    7: "U40",
    8: "i2",
    9: "f4",
    10: "u4",
    11: "b1",
    12: "i4",
    13: "f8",
}

dsize = {
    0: 40,
    1: 2,
    2: 4,
    3: 4,
    4: 1,
    5: 4,
    6: 8,
    7: 40,
    8: 2,
    9: 4,
    10: 4,
    11: 1,
    12: 4,
    13: 8,
}

dbrtypes = {
    "STRING": 0,
    "INT": 1,
    "SHORT": 1,
    "FLOAT": 2,
    "ENUM": 3,
    "CHAR": 4,
    "LONG": 5,
    "DOUBLE": 6,
}


class Chunk(object):
    def __init__(self, header):
        self.header = header
        self.events = []
        self.meta = None
        self.value = None

    def add(self, event):
        self.events.append(event)



class DefaultArchiverNotFound(Exception):
    """Default archiver was not found.

    Typically raised:

    The configuration file is read. In the default section the
    default archiver is named but no section in the configuration
    file uses the same name.
    """


class ArchiverReturnedNoData(Exception):
    """Arichver returned no data for valid request

    Typically raised:
          Request is accepted but no data were returned.
    """



# -*- coding: utf-8 -*-
# Author: Andreas Schaelicke <andreas.schaelicke@helmholtz-berlin.de>
#         Pierre Schnizer <pierre.schnizer@helmholtz-berlin.de>
# Time-stamp:
"""Archiver interface pythonic version

"""
from .archiver import ArchiverBasis
from .protocol_buffer import Chunk, dtypes as _dtypes, decoder as _decoder
from . import epics_event_pb2 as proto

from urllib.request import urlopen, quote
import numpy as np
import logging

logger = logging.getLogger('bact_archiver')


class Archiver(ArchiverBasis):
    def _getData(self, pvname, *, t0, t1, **kwargs):
        fmt = self.data_url_fmt
        url = fmt.format(format='raw', var=quote(pvname),
                         t0=quote(t0), t1=quote(t1))
        f = urlopen(url)
        try:
            data = get_data(f.read())
        except Exception as exc:
            fmt = 'Data retrieval using url {} raised exception {}'
            logger.error(fmt.format(url, exc))
            raise

        return data


def get_data(data):
    chunks = []

    for seq in data.split(b'\n\n'):
        lines = seq.splitlines()

        header = proto.PayloadInfo()
        header.ParseFromString(
            lines.pop(0).replace(b'\x1b\x02', b'\x0a').replace(
                b'\x1b\x03', b'\x0d').replace(b'\x1b\x01', b'\x1b'))
        chunk = Chunk(header)
        chunks.append(chunk)

        dtype = _dtypes[header.type]

        if 7 <= header.type <= 13:
            # vector decoder
            def parse_vec(inp, event=_decoder[header.type]()):
                event.ParseFromString(
                    inp.replace(b'\x1b\x02', b'\x0a').replace(
                        b'\x1b\x03', b'\x0d').replace(b'\x1b\x01', b'\x1b'))
                return [v
                        for v in event.val], event.secondsintoyear, event.nano

            # Not sure if the same applied for scalars. So for now I put it here
            if len(lines) == 0:
                fmt = 'Archiver did not return data. header = %s'
                logger.info(fmt, header)
                # raise ArchiverReturnedNoData(txt)
                chunk.values = None
                continue

            tmp = [parse_vec(l) for l in lines]
            chunk.values = np.array(tmp,
                dtype=[('value', dtype, (header.elementCount, )),
                       ('sec', 'u4'), ('ns', 'u4')])
        else:
            # scalar decoder
            def parse(inp, event=_decoder[header.type]()):
                event.ParseFromString(
                    inp.replace(b'\x1b\x02', b'\x0a').replace(
                        b'\x1b\x03', b'\x0d').replace(b'\x1b\x01', b'\x1b'))
                return event.val, event.secondsintoyear, event.nano

            chunk.values = np.array(
                [parse(l) for l in lines],
                dtype=[('value', dtype), ('sec', 'u4'), ('ns', 'u4')])

    # Todo:
    #     Check if all values were received or if one is misisng
    vals = [c.values for c in chunks if c.values is not None]
    arr =  np.concatenate(vals)
    return header, arr



# Author: Pierre Schnizer <pierre.schnizer@helmholtz-berlin.de>
# Date: Sep. 2017
from __future__ import absolute_import, print_function, division
import os.path

_dir_name = os.path.dirname(__file__)
test_data_dir = os.path.join(_dir_name, "test_data")
test_data_dir = os.path.normpath(test_data_dir)



from typing import Sequence

import dateutil
import pandas as pd

def convert_data_time(years: Sequence, secs: Sequence, nsecs: Sequence) -> pd.DatetimeIndex:
    """convert time given in years, secs and nsecs to DatetimeIndex
    Todo:
        use the function in the archiver
    """
    df = pd.DataFrame(
        {"year": years, "month": 1, "day": 1, "second": secs, "ns": nsecs}
    )
    dt = pd.to_datetime(df, utc=True)
    dt = pd.DatetimeIndex(dt, name="datetime").tz_convert(dateutil.tz.tzlocal())
    return dt



from bact_archiver.utils import convert_data_time
import numpy as np


def test_convert_date_time():
    import datetime
    import dateutil
    import pytz

    now = datetime.datetime(2023, 9, 12, 9, 42, 45)
    a_day = datetime.timedelta(seconds=24 * 60 * 60)

    start = now - 2 * a_day
    # every 10 minutes
    step = a_day / 24 / 6
    points = [start + step * i for i in range(10)]
    year_start = datetime.datetime(start.year, month=1, day=1)

    seconds = [int((p - year_start).total_seconds()) for p in points]
    nsecs = [v * 10 + 100 for v in range(len(seconds))]
    print(points)
    dt = convert_data_time([start.year] * len(seconds), seconds, nsecs)

    # need to add points their TimeZone: UTC
    time_deltas = [t_p.astimezone(pytz.utc) - p.replace(tzinfo=pytz.utc) for p, t_p in zip(points, dt)]

    for dt, t_nsec in zip(time_deltas, nsecs):
        assert dt == np.timedelta64(t_nsec, "ns")



import sphinx_rtd_theme


extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.napoleon",
    "sphinx.ext.intersphinx",
    "sphinx.ext.imgmath",
    "sphinx.ext.todo",
    "sphinx.ext.viewcode",
]

# Napoleon settings
napoleon_google_docstring = True
napoleon_numpy_docstring = True
napoleon_include_init_with_doc = True
napoleon_include_private_with_doc = True
napoleon_include_special_with_doc = True
napoleon_use_admonition_for_examples = False
napoleon_use_admonition_for_notes = False
napoleon_use_admonition_for_references = False
napoleon_use_ivar = True
napoleon_use_param = True
napoleon_use_rtype = True

# Intersphinx mapping
intersphinx_mapping = {
    "python": ("https://docs.python.org/3.7", None),
    "h5py": ("http://docs.h5py.org/en/stable/", None),
    "pandas": ("https://pandas.pydata.org/docs/", None),
}

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
# source_suffix = ['.rst', '.md']
source_suffix = ".rst"

# The master toctree document.
master_doc = "index"

# General information about the project.
project = "bact-archiver"
copyright = 'Helmholtzzentrum Berlin f"ur Materialien und Energie, 2019'
author = 'Andreas Sch"alike, Pierre Schnizer'
# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The short X.Y version.
version = "0.0"
# The full version, including alpha/beta/rc tags.
release = "beta"

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = "en"

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This patterns also effect to html_static_path and html_extra_path
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = True


# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
# html_theme = 'alabaster'
html_theme = "sphinx_rtd_theme"
html_theme_path = [sphinx_rtd_theme.get_html_theme_path()]

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#

# html_theme_options = {}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
# html_static_path = ['_static']

# Custom sidebar templates, must be a dictionary that maps document names
# to template names.
#
# This is required for the alabaster theme
# refs: http://alabaster.readthedocs.io/en/latest/installation.html#sidebars
html_sidebars = {
    "**": [
        "about.html",
        "navigation.html",
        "relations.html",  # needs 'show_related': True theme option to display
        "searchbox.html",
        "donate.html",
    ]
}


# -- Options for HTMLHelp output ------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = "bactdoc"


# -- Options for LaTeX output ---------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    #
    # 'papersize': 'letterpaper',
    "papersize": "a4paper",
    # The font size ('10pt', '11pt' or '12pt').
    #
    # 'pointsize': '10pt',
    # Additional stuff for the LaTeX preamble.
    #
    # 'preamble': '',
    # Latex figure (float) alignment
    #
    # 'figure_align': 'htbp',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (master_doc, "bact2.tex", "bact Documentation", "Author", "manual"),
]


# -- Options for manual page output ---------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [(master_doc, "bact2", "bact2 Documentation", [author], 1)]


# -- Options for Texinfo output -------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (
        master_doc,
        "bact2",
        "bact2 Documentation",
        author,
        "bact2",
        "Commissioning tools for accelerators",
        "Miscellaneous",
    ),
]


# -- Options for Epub output ----------------------------------------------

# Bibliographic Dublin Core info.
epub_title = project
epub_author = author
epub_publisher = author
epub_copyright = copyright

# The unique identifier of the text. This can be a ISBN number
# or the project homepage.
#
# epub_identifier = ''

# A unique identification for the text.
#
# epub_uid = ''

# A list of files that should not be packed into the epub file.
epub_exclude_files = ["search.html"]



import unittest
import bact_archiver.config


class LoadConfig(unittest.TestCase):

    mod_name = "bact_archiver"

    def test001_LoadConfigurationFile(self):
        """Test that the archiver configuration file is fonnd"""
        filename = bact_archiver.config.get_config_filename(self.mod_name)
        self.assertIsNotNone(filename)

    def test002_LoadConfiguration(self):
        """Test that the archiver configurations are loaded"""
        bact_archiver.config.archiver_configurations(self.mod_name)

    def test003_LoadConfiguration(self):
        """Test that the archiver configurations have a name"""
        tmp = bact_archiver.config.archiver_configurations(self.mod_name)
        for name, archiver in tmp.items():
            self.assertIsNotNone(archiver.name)

    def test004_LoadConfiguration(self):
        """Test that the archiver configurations are printable"""
        tmp = bact_archiver.config.archiver_configurations(self.mod_name)
        for name, archiver in tmp.items():
            str(archiver)
            repr(archiver)

    def test010_ModuleInsertConfigurations(self):
        """Test that the archiver configurations are inserted as archivers in
        the archiver module
        """

        import bact_archiver.archiver
        import bact_archiver.register

        configs = bact_archiver.config.archiver_configurations(self.mod_name)
        bact_archiver.register.add_archivers_to_module(self.mod_name, configs)

    def test020_AutomaticLoadDefault(self):
        import bact_archiver

        default_archiver = bact_archiver.default
        assert default_archiver

    def test021_AutomaticLoadDefaultName(self):
        import bact_archiver

        default_archiver = bact_archiver.default
        default_archiver.name

    def test022_AutomaticLoadDefaultName(self):
        import bact_archiver

        default_archiver = bact_archiver.default
        default_archiver.description


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
# Authors : Andreas Schälicke <andreas.schaelike@helmholtz-berlin.de>
#           Pierre Schnizer <pierre.schnizer@helmholtz-berlin.de>
# Date : 2017, 2020, 2023, 2024
import os
import sys

from setuptools import Extension, setup
from numpy import get_include
# required that protocol_buffer is found
sys.path.append(os.path.dirname(__file__))
from protocol_buffer import GenerateProtocolBuffer

setup(
    cmdclass=dict(build_proto_c=GenerateProtocolBuffer),
    ext_modules=[
        Extension(
            name="bact_archiver.epics_event",
            sources=[
                "proto/epics_event.pyx",
                "proto_gen/epics_event.pb.cc",
            ],
            include_dirs=[".", "proto/", "proto_gen/", get_include()],
            libraries=[
                "protobuf",
            ],
            language="c++",
        )
    ],
)



import os
import unittest
import logging

from bact_archiver.pyarchiver import get_data
from common import test_data_dir

log = logging.getLogger('test')


class ArchiverClientTest(unittest.TestCase):
    """Test the decoding of PB/HTTP formated archiver data
    """

    def read(self, fname):
        fullname = os.path.join(test_data_dir, fname)
        log.info('reading {}'.format(fullname))
        with open(fullname, 'rb') as f:
            return f.read()

    def run_decode(self, *, fname=None, **kw):
        data = self.read(fname)

        header, values = get_data(data)
        log.debug('Header %s', header)

        self.assertEqual(header.type, kw['exp_type'])
        self.assertEqual(len(values), kw['exp_len'])

        self.assertEqual(header.elementCount, kw.get('exp_count', 1))

    def test00_scalar_string(self):

        self.run_decode(fname='20171101_stGun.pb',
                        exp_type=0,
                        exp_len=5)

    def test03_scalar_enum(self):

        self.run_decode(fname='20171101_selTrgSR.pb',
                        exp_type=3,
                        exp_len=4)
        # TODO check all these nice extra info in the header ...

    def test05_scalar_long(self):
        self.run_decode(fname='20171101_numShots.pb',
                        exp_type=5,
                        exp_len=29)

    def test06_scalar_double(self):

        self.run_decode(fname='201710010200_rdCur.pb',
                        exp_type=6,
                        exp_len=56)

    def test08_vector_short(self):

        self.run_decode(fname='20171101_sram_maxrms.pb',
                        exp_type=8,
                        exp_len=121,
                        exp_count=31457)

    def test09_vector_float(self):

        self.run_decode(fname='20171101_sram_mean.pb',
                        exp_type=9,
                        exp_len=241,
                        exp_count=400)

    @unittest.skip('broken input file')
    def test11_vector_char(self):

        self.run_decode(fname='20171101_sram_mask.pb',
                        exp_type=11,
                        exp_len=121,
                        exp_count=400)

    @unittest.skip('broken input file')
    def test12_vector_int(self):

        self.run_decode(fname='20171101_fullhistogram.pb',
                        exp_type=12,
                        exp_len=121,
                        exp_count=6255)

    def test13_vector_double(self):

        self.run_decode(fname='20171101_MBcurrent.pb',
                        exp_type=13,
                        exp_len=121,
                        exp_count=400)


if __name__ == "__main__":
    unittest.main()
